-- name: Climb Attack
-- description: Players compete to climb a randomly generated tower using block building abilities! The player who climbs the highest wins!\n\nCreator: eeliottheking\n\nCredits:\nMusic by Eric Matyas - www.soundimage.org\nMap (Roblox Baseplate) by beardenthusiast\nMinecraft mod by zKevin\n\nAnd thank you to all the publicly available mods on the github for information and resources!
-- incompatible: gamemode

-- Localize functions to improve performance
local
hook_chat_command, network_player_set_description, network_is_server,
hook_event, play_sound, djui_chat_message_create, djui_hud_set_resolution, djui_hud_set_font,
djui_hud_set_color, djui_hud_render_rect, djui_hud_print_text, djui_hud_get_screen_width, djui_hud_get_screen_height,
djui_hud_measure_text, warp_to_start_level, dist_between_objects,
math_floor, audio_stream_load, audio_stream_set_looping, audio_stream_play,
obj_get_nearest_object_with_behavior_id, spawn_sync_object, spawn_non_sync_object, obj_get_first, obj_get_next, get_id_from_behavior, cur_obj_scale
=
hook_chat_command, network_player_set_description, network_is_server,
hook_event, play_sound, djui_chat_message_create, djui_hud_set_resolution, djui_hud_set_font,
djui_hud_set_color, djui_hud_render_rect, djui_hud_print_text, djui_hud_get_screen_width, djui_hud_get_screen_height,
djui_hud_measure_text, warp_to_start_level, dist_between_objects,
math.floor, audio_stream_load, audio_stream_set_looping, audio_stream_play,
obj_get_nearest_object_with_behavior_id, spawn_sync_object, spawn_non_sync_object, obj_get_first, obj_get_next, get_id_from_behavior, cur_obj_scale

-- Constants
ROUND_WAITING_PLAYERS = 0
ROUND_WAITING = 1
ROUND_ACTIVE = 2
ROUND_CLEANUP = 3
GRID_SIZE = 200
MAX_HEALTH = 0x8C0

-- Globals
gGlobalSyncTable.roundState = ROUND_WAITING_PLAYERS -- Round state
gGlobalSyncTable.warpPlayersToOrigin = false
gGlobalSyncTable.displayTimer = 0
gGlobalSyncTable.winIndex = -1
gGlobalSyncTable.numBoxesToSpawn = 100 -- Default amount of boxes to spawn each round
gGlobalSyncTable.maxBoxesAllowed = 10 -- Default amount of boxes players can build each round
gGlobalSyncTable.matchDuration = 90 -- Default match duration in seconds
gGlobalSyncTable.boxVerticalSpawnRange = 50 -- Default vertical spawn range for boxes (measured in boxes)
gGlobalSyncTable.boxHorizontalRange = 5 -- Default horizontal spawn range for boxes (measured in boxes).  For z and x axis (5 is a 5x5 square)
gGlobalSyncTable.cleanupTimer = 10 -- Timer for my hacky level clear function.

-- Create tables to track player's blocks placed an their block break timer and start them at 0
for i = 0, (MAX_PLAYERS - 1) do
    gPlayerSyncTable[i].blocksPlaced = 0
    gPlayerSyncTable[i].blockBreakTimer = 0
end

-- Additional variables
local startTimer = 15 * 30 + 30 -- 15 seconds
local activeTimer = gGlobalSyncTable.matchDuration * 30 + 30 -- Match duration in seconds
local playerHeights = {} -- Table to store player heights
local audioStream = nil -- Variable for our audio stream
local showRulesPopup = true -- Variable to check if the rules have been closed

-- function to display the rules pop-up
local function display_rules_popup(m)
    if m.playerIndex ~= 0 then return end

    djui_hud_set_resolution(RESOLUTION_DJUI)
    djui_hud_set_font(FONT_NORMAL)

    local screenWidth = djui_hud_get_screen_width()
    local screenHeight = djui_hud_get_screen_height()

    local popupWidth = 1000
    local popupHeight = 600
    local popupX = (screenWidth - popupWidth) / 2
    local popupY = (screenHeight - popupHeight) / 2

    -- Render the background rectangle
    djui_hud_set_color(0, 0, 0, 170)
    djui_hud_render_rect(popupX, popupY, popupWidth, popupHeight)

    -- Render the title
    djui_hud_set_color(255, 255, 255, 255)
    djui_hud_print_text("Climb Attack Rules", popupX + 20, popupY + 20, 1.5)

    -- Render the rules
    local rulesText = {
        "THIS MOD IS NOT OPTIMISED FOR MOBILE!",
        "1. Players compete to climb the highest using block building abilities.",
        "2. A new tower spawns each round, the player who climbs the highest wins!",
        "3. Use the Y button to place blocks and build your way up.",
        "4. You have a limited number of blocks to place.",
        "5. Punching or kicking blocks will refund your block count.",
        "6. Have fun and good luck!",
        "Music by Eric Matyas - www.soundimage.org",
        "Map (Roblox Baseplate) by beardenthusiast",
        "Minecraft mod by zKevin"
    }

    local lineHeight = 30
    local y = popupY + 100
    for _, line in ipairs(rulesText) do
        djui_hud_print_text(line, popupX + 20, y, 1.2)
        y = y + lineHeight
    end

    -- Render the "Press A or jump to continue" text
    djui_hud_set_color(255, 255, 0, 255)
    djui_hud_print_text("Press A or Jump to continue", popupX + 20, popupY + popupHeight - 80, 1.2)

    -- Disable input for Mario until the user presses the A button
    if m.controller.buttonPressed & A_BUTTON ~= 0 then
        showRulesPopup = false
        set_mario_action(m, ACT_IDLE, 0)
    else
        set_mario_action(m, ACT_READING_AUTOMATIC_DIALOG, 0)
    end
end

-- Function to only show the rules popup for the local player
local function on_player_connected(m)
    if m.playerIndex == 0 then
        showRulesPopup = true
    end
end

-- Hook the rules
hook_event(HOOK_ON_PLAYER_CONNECTED, on_player_connected)
hook_event(HOOK_ON_HUD_RENDER, function()
    if showRulesPopup then
        display_rules_popup(gMarioStates[0])
    end
end)

-- Funtion to give mario infinite health
local function give_infinite_health(m)
    -- Check if the player's health is below the maximum
    if m.health < MAX_HEALTH then
        -- Set the player's health to the maximum value
        m.health = MAX_HEALTH
    end
end

-- Function to play custom music
function on_level_init()
    audioStream = audio_stream_load("background.mp3")
    audio_stream_set_looping(audioStream, true)
    audio_stream_play(audioStream, true, 1)
end

-- Hook custom music to level initialization
hook_event(HOOK_ON_LEVEL_INIT, on_level_init)

-- Main server functions
function server_update()

    -- Count how many players are connected
    local connectedCount = 0
    for i = 0, (MAX_PLAYERS - 1) do
        if gNetworkPlayers[i].connected then
            connectedCount = connectedCount + 1
        end
    end

    -- Set round state to ROUND_WAITING_PLAYERS if there aren't enough players
    if connectedCount < 2 then
        gGlobalSyncTable.roundState = ROUND_WAITING_PLAYERS
        gGlobalSyncTable.winIndex = -1
        startTimer = 15 * 30 + 30 -- Reset startTimer
        return
    end

    -- Set state to ROUND_WAITING when enough players are connected to alert everyone that the match is starting
    if connectedCount > 1 and gGlobalSyncTable.roundState == ROUND_WAITING_PLAYERS then
        gGlobalSyncTable.roundState = ROUND_WAITING
        startTimer = 15 * 30 + 30 -- Reset startTimer
    end

    -- If state is ROUND_WAITING...
    if gGlobalSyncTable.roundState == ROUND_WAITING then
        startTimer = startTimer - 1 -- Countdown timer
        gGlobalSyncTable.displayTimer = startTimer -- Use startTimer for displayTimer
        if startTimer <= 0 then  -- Our start timer has ended
            activeTimer = gGlobalSyncTable.matchDuration * 30 + 30 -- Set match timer to specified time
            gGlobalSyncTable.roundState = ROUND_ACTIVE -- Transition to ROUND_ACTIVE
            spawn_random_boxes() -- Spawn boxes when the round starts
            startTimer = 15 * 30 + 30 -- Reset startTimer for next iteration
            -- Reset blocksPlaced count for each player
            for i = 0, (MAX_PLAYERS - 1) do
                gPlayerSyncTable[i].blocksPlaced = 0
            end
        end
    end

    -- If state is ROUND_ACTIVE...
    if gGlobalSyncTable.roundState == ROUND_ACTIVE then
        activeTimer = activeTimer - 1 -- Countdown timer
        gGlobalSyncTable.displayTimer = activeTimer -- Make our display timer the game timer
        gGlobalSyncTable.winIndex = -1 -- Reset stored winning player

        -- Find the player with the highest y-height
        local highestPlayer = -1
        local highestY = -20000
        for i = 0, (MAX_PLAYERS - 1) do
            if gNetworkPlayers[i].connected and playerHeights[i] ~= nil and playerHeights[i] > highestY then
                highestY = playerHeights[i]
                highestPlayer = i
            end
        end
        
        -- Store winning player and transition into ROUND_CLEANUP
        if activeTimer <= 0 then
            gGlobalSyncTable.winIndex = highestPlayer
            gGlobalSyncTable.roundState = ROUND_CLEANUP
        end
    end

    -- if state is ROUND_CLEANUP...
    if gGlobalSyncTable.roundState == ROUND_CLEANUP then
        gGlobalSyncTable.warpPlayersToOrigin = true -- Enable our flag that clears the level
        gGlobalSyncTable.cleanupTimer = gGlobalSyncTable.cleanupTimer - 1 -- Countdown the cleanupTimer

        -- When the cleanup timer reaches 0, transition back to ROUND_WAITING_PLAYERS
        if gGlobalSyncTable.cleanupTimer <= 0 then
            gGlobalSyncTable.warpPlayersToOrigin = false -- Reset the warp flag here
            gGlobalSyncTable.roundState = ROUND_WAITING_PLAYERS
            gGlobalSyncTable.cleanupTimer = 10 -- Reset the cleanupTimer
        end
    end
end

-- Our main mario function
function mario_update(m)

    -- If the player is the local player do everything in mario_update_local
    if m.playerIndex == 0 then
        mario_update_local(m)
    end
    give_infinite_health(m) -- Give mario infinite health
    playerHeights[m.playerIndex] = m.pos.y -- Update player height in the playerHeights table

    -- Set player description with height from the playerHeights table
    local height = math_floor(playerHeights[m.playerIndex]) - 50
    network_player_set_description(gNetworkPlayers[m.playerIndex], "H: " .. height, 255, 255, 255, 255)

    -- Increment the block break timer for each player
    local sMario = gPlayerSyncTable[m.playerIndex]
    sMario.blockBreakTimer = sMario.blockBreakTimer + 1

    -- Check if our warping flag has been set, warping all the marios to reload the level and clear the blocks
    if gGlobalSyncTable.warpPlayersToOrigin then
        warp_to_start_level()
    end
end

-- Our mario update function for just the local player
function mario_update_local(m)

    -- Give player block building abilities from Minecraft mod
    local rgt = math.sin(m.intendedYaw / 32688 * math.pi)
    local fwd = math.cos(m.intendedYaw / 32688 * math.pi)
    local in_air = m.pos.y - m.floorHeight > 5
    local crouching = (m.controller.buttonDown & Z_TRIG) ~= 0
    local posX = to_grid(m.pos.x + (in_air and 0 or rgt * 200))
    local posY = to_grid(m.pos.y + ((in_air or crouching) and -200 or 0))
    local posZ = to_grid(m.pos.z + (in_air and 0 or fwd * 200))

    -- Only allow block placement during ROUND_ACTIVE
    if gGlobalSyncTable.roundState == ROUND_ACTIVE then

        -- Place block
        if (m.controller.buttonPressed & Y_BUTTON) ~= 0 then
            local nearest = obj_get_nearest_object_with_behavior_id(m.marioObj, id_bhvBreakableBox)

            if nearest then
                local dist = dist_between_objects(m.marioObj, nearest)
                if dist >= 190 then
                    place_block(m, posX, posY, posZ)
                else
                    play_sound(SOUND_MENU_CAMERA_BUZZ, m.marioObj.header.gfx.cameraToObject)
                end
            else
                place_block(m, posX, posY, posZ)
            end
        end
    end

    -- Outline block
    local place = find_place()
    if not place then
        place = spawn_non_sync_object(
            id_bhvOutlineblock,
            E_MODEL_EXCLAMATION_BOX_OUTLINE,
            posX, posY, posZ,
            nil
        )
    else
        place.oPosX = posX
        place.oPosY = posY - 8
        place.oPosZ = posZ
    end
end

-- Check if the player is the server before running code in server_update
function update()
    if network_is_server() then
        server_update()
    end
end

-- Render top HUD based on round state
function hud_top_render()
    local text = ''
    if gGlobalSyncTable.roundState == ROUND_WAITING_PLAYERS then
        text = 'Waiting for Players'
    elseif gGlobalSyncTable.roundState == ROUND_ACTIVE then
        text = 'Time Remaining: ' .. math_floor(gGlobalSyncTable.displayTimer / 30)
    elseif gGlobalSyncTable.roundState == ROUND_WAITING then
        text = 'Next round in ' .. math_floor(gGlobalSyncTable.displayTimer / 30)
    else
        return false
    end

    local scale = 0.50
    local screenWidth = djui_hud_get_screen_width()
    local width = djui_hud_measure_text(text) * scale
    local x = (screenWidth - width) / 2.0
    local y = 0

    djui_hud_set_color(0, 0, 0, 128)
    djui_hud_render_rect(x - 6, y, width + 12, 16)
    djui_hud_set_color(255, 255, 255, 255)
    djui_hud_print_text(text, x, y, scale)
end

-- Show winner popup
function hud_center_render()
    if gGlobalSyncTable.winIndex < 0 or gGlobalSyncTable.winIndex > 16 then
        return
    end

    local text = gNetworkPlayers[network_local_index_from_global(gGlobalSyncTable.winIndex)].name .. " Wins!"
    local scale = 1
    local screenWidth = djui_hud_get_screen_width()
    local screenHeight = djui_hud_get_screen_height()
    local width = djui_hud_measure_text(text) * scale
    local height = 32 * scale
    local x = (screenWidth - width) / 2.0
    local y = (screenHeight - height) / 2.0

    djui_hud_set_color(0, 0, 0, 128)
    djui_hud_render_rect(x - 6 * scale, y, width + 12 * scale, height)
    djui_hud_set_color(255, 255, 255, 255)
    djui_hud_print_text(text, x, y, scale)
end

-- Hud element for keeping track of how many boxes each player has left
function hud_bottom_left_render()
    local m = gMarioStates[0]
    local sMario = gPlayerSyncTable[m.playerIndex]
    local blocksLeft = gGlobalSyncTable.maxBoxesAllowed - sMario.blocksPlaced

    local text = "Blocks Left: " .. blocksLeft
    local scale = 0.5
    local screenWidth = djui_hud_get_screen_width()
    local width = djui_hud_measure_text(text) * scale
    local x = 10
    local y = djui_hud_get_screen_height() - 30 * scale

    -- Render background
    djui_hud_set_color(0, 0, 0, 128)
    djui_hud_render_rect(x - 6, y - 6, width + 12, 16)

    -- Render text
    djui_hud_set_color(255, 255, 255, 255)
    djui_hud_print_text(text, x, y - 6, scale)
end

-- Function to render the HUD
function on_hud_render()
    hud_hide()
    djui_hud_set_font(FONT_NORMAL)
    djui_hud_set_resolution(RESOLUTION_N64)
    hud_top_render()
    hud_center_render()
    hud_bottom_left_render()
end

-- This is my hacky way of checking if mario has broken a block and giving him another block to build
function on_interact(m, obj, interactType)
    if interactType == INTERACT_BREAKABLE and m.action == ACT_PUNCHING or m.action == ACT_MOVE_PUNCHING or m.action == ACT_JUMP_KICK then
        local sMario = gPlayerSyncTable[m.playerIndex]
        if sMario.blocksPlaced > 0 and sMario.blockBreakTimer >= 8 then
            sMario.blocksPlaced = sMario.blocksPlaced - 1
            sMario.blockBreakTimer = 0
        end
    end
end

-- Hooks
hook_event(HOOK_UPDATE, update)
hook_event(HOOK_MARIO_UPDATE, mario_update)
hook_event(HOOK_ON_HUD_RENDER, on_hud_render)
hook_event(HOOK_ON_INTERACT, on_interact)

-- Minecraft mod functions
function to_grid(n)
    return math_floor(n / GRID_SIZE + .5) * GRID_SIZE
end

function bhv_outlineblock_init(obj)
    obj.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
    cur_obj_scale(4.01)
    obj.oOpacity = 255
    obj.oFaceAnglePitch = 0
    obj.oFaceAngleYaw = 0
    obj.oFaceAngleRoll = 0
end

id_bhvOutlineblock = hook_behavior(nil, OBJ_LIST_DEFAULT, true, bhv_outlineblock_init, nil)

function find_place()
    local obj = obj_get_first(OBJ_LIST_DEFAULT)
    while obj ~= nil do
        if get_id_from_behavior(obj.behavior) == id_bhvOutlineblock then
            return obj
        end
        obj = obj_get_next(obj)
    end
    return nil
end

function place_block(m, x, y, z)
    local sMario = gPlayerSyncTable[m.playerIndex]
    if sMario.blocksPlaced < gGlobalSyncTable.maxBoxesAllowed then
        local box = spawn_sync_object(
            id_bhvBreakableBox,
            E_MODEL_BREAKABLE_BOX,
            x, y, z,
            function(obj)
                obj.oOpacity = 255
                obj.oFaceAnglePitch = 0
                obj.oFaceAngleYaw = 0
                obj.oFaceAngleRoll = 0
            end
        )

        play_sound(SOUND_GENERAL_BOX_LANDING, { x = x, y = y, z = z })
        sMario.blocksPlaced = sMario.blocksPlaced + 1
    else
        play_sound(SOUND_MENU_CAMERA_BUZZ, m.marioObj.header.gfx.cameraToObject)
    end
end

-- Function to build the tower
function spawn_random_boxes()
    local numBoxes = gGlobalSyncTable.numBoxesToSpawn -- Get the amount of boxes to spawn
    local startX, startZ = 0, 0 -- Starting position for the boxes
    local verticalRange = gGlobalSyncTable.boxVerticalSpawnRange * 100 -- Convert to units
    local horizontalRange = gGlobalSyncTable.boxHorizontalRange * 100 -- Use the new horizontal range value
    local minDistanceFromOrigin = 100 -- Minimum distance from the origin (0, 0, 0) in units (100 is one block)
    local minYOffset = 50 -- Minimum Y offset from the origin in units (Account for 'nub' on baseplate map)

    for i = 1, numBoxes do
        local x, z
        repeat
            x = to_grid(startX + math.random(-horizontalRange, horizontalRange))
            z = to_grid(startZ + math.random(-horizontalRange, horizontalRange))
        until math.sqrt(x * x + z * z) >= minDistanceFromOrigin -- Check distance from origin

        local y
        repeat
            y = to_grid(math.random(minYOffset, verticalRange))
        until y >= minYOffset -- Ensure Y offset is at least minYOffset

        spawn_sync_object(
            id_bhvBreakableBox,
            E_MODEL_BREAKABLE_BOX,
            x, y, z,
            function(obj)
                obj.oOpacity = 255
                obj.oFaceAnglePitch = 0
                obj.oFaceAngleYaw = 0
                obj.oFaceAngleRoll = 0
            end
        )
    end
end

-- Function for command to set the number of boxes to spawn
local function on_set_num_boxes_command(msg)
    local num = tonumber(msg)
    if num and num >= 0 and num <= 200 then
        gGlobalSyncTable.numBoxesToSpawn = num
        djui_chat_message_create("Number of boxes to spawn set to " .. num)
    else
        djui_chat_message_create("Invalid number. Please enter a value between 0 and 200.")
    end
    return true
end

-- Function to set the match length
local function on_set_match_duration_command(msg)
    local duration = tonumber(msg)
    if duration and duration >= 30 and duration <= 300 then
        gGlobalSyncTable.matchDuration = duration
        djui_chat_message_create("Match duration set to " .. duration .. " seconds")
    else
        djui_chat_message_create("Invalid duration. Please enter a value between 30 and 300 seconds.")
    end
    return true
end

-- Function to set how many boxes each player can build
local function on_set_player_boxes_command(msg)
    local num = tonumber(msg)
    if num and num >= 0 and num <= 30 then
        gGlobalSyncTable.maxBoxesAllowed = num
        djui_chat_message_create("Maximum number of boxes allowed per player set to " .. num)
    else
        djui_chat_message_create("Invalid number. Please enter a value between 0 and 30.")
    end
    return true
end

-- Function for a chat command to set the horizontal ranges for the block tower
local function on_set_box_horizontal_range_command(msg)
    local range = tonumber(msg)
    if range and range >= 3 and range <= 20 then
        gGlobalSyncTable.boxHorizontalRange = range
        djui_chat_message_create("Horizontal spawn range for boxes set to " .. range)
    else
        djui_chat_message_create("Invalid range. Please enter a value between 3 and 20.")
    end
    return true
end

-- Function to set vertical height limit of tower
local function on_set_box_vertical_range_command(msg)
    local range = tonumber(msg)
    if range and range >= 1 and range <= 300 then
        gGlobalSyncTable.boxVerticalSpawnRange = range
        djui_chat_message_create("Vertical spawn range for boxes set to " .. range)
    else
        djui_chat_message_create("Invalid range. Please enter a value between 1 and 300.")
    end
    return true
end

-- Function to force the map to restart
local function on_round_command(msg)
    if msg == "restart" then
        if gGlobalSyncTable.roundState == ROUND_ACTIVE then
            -- Start a new cleanup round
            gGlobalSyncTable.winIndex = -1
            gGlobalSyncTable.roundState = ROUND_CLEANUP
            djui_chat_message_create("Round restarted!")
        else
            djui_chat_message_create("Cannot restart right now.")
        end
    else
        djui_chat_message_create("Invalid command. Use '/round restart'.")
    end
    return true
end

-- Hook server commands
if network_is_server() then
    hook_chat_command("round", "('restart')' Start a new round.", on_round_command)
    hook_chat_command("set-num-boxes", "(number) Set the number of boxes to spawn. Default: 100", on_set_num_boxes_command)
    hook_chat_command("set-match-duration", "(number) Set the match duration in seconds. Default: 90", on_set_match_duration_command)
    hook_chat_command("set-box-vertical-range", "(number) Set the vertical spawn range for boxes in box units. Default: 50", on_set_box_vertical_range_command)
    hook_chat_command("set-box-horizontal-range", "(number) Set the horizontal spawn range for boxes in box units. Default: 5", on_set_box_horizontal_range_command)
    hook_chat_command("set-player-boxes", "(number) Set the maximum number of boxes allowed for players. Default: 10", on_set_player_boxes_command) -- New chat command
end